# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: Dec 26, 2024

# Description: This script contains the implementations of the IPW and doubly 
# robust estimators of the derivative of a dose-response curve with and without 
# the positivity condition.

import numpy as np
from .rbf import KernelRetrieval
from .utils import CondDenEst, CondDenEstKDE, KDE
from sklearn.model_selection import KFold
from sklearn.base import BaseEstimator

import torch
import torch.nn as nn
import torch.optim as optim

#=======================================================================================#
# Implementations of the proposed estimators that assume the positivity condition

## Define the neural network
class NeurNet(nn.Module):
    def __init__(self, input_size):
        super(NeurNet, self).__init__()
        self.fc1 = nn.Linear(input_size, 100)  # First layer
        self.silu = nn.SiLU()
        self.fc2 = nn.Linear(100, 50) # Second layer
        self.fc3 = nn.Linear(50, 1)
        
        # Apply Kaiming initialization to each layer
        nn.init.kaiming_normal_(self.fc1.weight, nonlinearity='relu')
        nn.init.kaiming_normal_(self.fc2.weight, nonlinearity='relu')
        nn.init.kaiming_normal_(self.fc3.weight, nonlinearity='linear')
        
        self.double()

    def forward(self, x):
        x = self.silu(self.fc1(x))
        x = self.silu(self.fc2(x))
        x = self.fc3(x)
        return x
    

def train(mod, X_train, Y_train, lr=0.01, n_epochs=10, momentum=0.7, weight_decay=0,
          print_loss=True):
    '''
    Utility function for training the PyTorch neural network model via stochastic
    gradient descent.
    
    Parameters
    ----------
        mod: python class
            The neural network class defined by PyTorch.
            
        X_train: (n,d+1)-torch.Tensor
            The first column of "X_train" is the treatment/exposure variable, 
            while the other d columns are the confounding variables of n observations.
            
        Y_train: (n,)-torch.Tensor
            The outcome variables of n observations.
            
        lr: float
            The learning rate (Default: lr=0.01.)
            
        n_epochs: int
            The number of training epochs. (Default: n_epochs=10.)
            
        momentum: float
            The momentum factor (Default: momentum=0.7.)
            
        weight_decay: float
            The weight decay (L2 penalty) (Default: weight_decay=0.)
            
        print_loss: boolean
            An indicator of whether the training loss will be printed to the console.
            
    Return
    ----------
        model: python object
            The fitted model instance of a neural network class defined by PyTorch.
    '''
    # Initialize the model, loss function, and optimizer
    model = mod(input_size=X_train.shape[1])
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    criterion = nn.MSELoss()    
    optimizer = optim.SGD(model.parameters(), lr=lr, momentum=momentum, 
                          weight_decay=weight_decay)
    
    for epoch in range(n_epochs):
        model.train()
        # Forward pass
        outputs = model(X_train)
        loss = criterion(outputs, Y_train)
        
        # Backward pass and optimization
        optimizer.zero_grad()    # Zero the gradients
        loss.backward()          # Backpropagate
        optimizer.step()         # Update weights
        if print_loss:
            print(f'Epoch [{epoch + 1}/{n_epochs}], Loss: {loss.item():.4f}')
        
    return model


def RADRDeriv(Y, X, t_eval, mu, L=1, n_iter=1000, lr=0.1, multi_boot=False, B=1000):
    '''
    Estimating the derivative of a dose-response curve through the regression 
    adjustment (or G-computation) form by a PyTorch neural network model under 
    the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        mu: a neural network class defined by PyTorch
            The conditional mean outcome (or regression) model of Y given X.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        n_iter: int
            The number of iterations or training epochs of the neural network model.
            (Default: n_iter=1000.)
            
        lr: float
            The learning rate (Default: lr=0.01.)
            
        multi_boot: boolean
            An indicator of whether the multiplier bootstrap will be run. 
            (Default: multi_boot=False.)
            
        B: int
            The number of bootstrapping times. (Default: B=1000.)
            
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
            
        mu_boot: (B,m)-array
            The estimated derivatives of the dose-response curves on bootstrapping 
            data evaluated at points "t_eval". (Only return this quantity when 
            "multi_boot=True".)
    '''
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the regression model on the entire data
        X_tensor = torch.from_numpy(X)
        Y_tensor = torch.from_numpy(Y.reshape(-1,1))
        NN_fit = train(mu, X_tensor, Y_tensor, lr=lr, n_epochs=n_iter)
        beta_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the fitted regression model
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            X_eval_tensor = torch.from_numpy(X_eval)
            for j in range(X_eval.shape[0]):
                # Compute the gradient of the fitted regression model with respect to the first coordinate
                x = X_eval_tensor[j,:]
                x = x.clone().detach().requires_grad_(True)
                y = NN_fit(x)
                y_scalar = y[0]
                y_scalar.backward()
                x_grad = x.grad
                beta_est[j,i] = x_grad[0].item()
            
        # theta_est = np.mean(beta_est, axis=0)
    else:
        # Conduct L-fold cross-fittings: fit the regression model on the training fold data
        # and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        beta_est = np.zeros((n, t_eval.shape[0]))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            
            X_tr_tensor = torch.from_numpy(X_tr)
            Y_tr_tensor = torch.from_numpy(Y_tr.reshape(-1,1))
            NN_fit = train(mu, X_tr_tensor, Y_tr_tensor, lr=lr, n_epochs=n_iter)
            for i in range(t_eval.shape[0]):
                # Define the data matrix for evaluating the fitted regression model
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                X_eval_te_tensor = torch.from_numpy(X_eval_te)
                beta_hat = np.zeros((X_eval_te.shape[0],))
                for j in range(X_eval_te.shape[0]):
                    # Compute the gradient of the fitted regression model with respect to the first coordinate
                    x = X_eval_te_tensor[j,:]
                    x = x.clone().detach().requires_grad_(True)
                    y = NN_fit(x)
                    y_scalar = y[0]
                    y_scalar.backward()
                    x_grad = x.grad
                    beta_hat[j] = x_grad[0].item()
                beta_est[te_ind,i] = beta_hat
                    
        # theta_est = np.mean(beta_est, axis=0)
    if multi_boot:
        theta_boot = np.zeros((B, t_eval.shape[0]))
        for b in range(B):
            Z = np.random.randn(n, t_eval.shape[0]) + 1
            theta_boot[b,:] = np.mean(Z * beta_est, axis=0)
        theta_est = np.mean(beta_est, axis=0)
        return theta_est, theta_boot
    else:
        theta_est = np.mean(beta_est, axis=0)
        return theta_est


def RADRDerivSKLearn(Y, X, t_eval, mu, L=1, delta=0.01):
    '''
    Estimating the derivative of a dose-response curve through the regression 
    adjustment (or G-computation) form under the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        mu: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        delta: float
            The step value for computing the finite differences (or numerical partial
            differentiation) of the fitted regression model.
            
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
    '''
    
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the regression model on the entire data
        mu_fit = mu.fit(X, Y)
        t_new = np.linspace(np.min(t_eval)-delta, np.max(t_eval)+delta, t_eval.shape[0]+1)
        beta_est = np.zeros((n, t_new.shape[0]))
        for i in range(t_new.shape[0]):
            # Define the data matrix for evaluating the fitted regression model
            X_eval = np.column_stack([t_new[i]*np.ones(n), X[:,1:]])
            beta_est[:,i] = mu_fit.predict(X_eval)
            
        theta_est = np.diff(np.mean(beta_est, axis=0))/np.diff(t_new)
    else:
        # Conduct L-fold cross-fittings: fit the regression model on the training fold data
        # and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        t_new = np.linspace(np.min(t_eval)-delta, np.max(t_eval)+delta, t_eval.shape[0]+1)
        beta_est = np.zeros((n, t_new.shape[0]))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            
            mu_fit = mu.fit(X_tr, Y_tr)
            for i in range(t_new.shape[0]):
                # Define the data matrix for evaluating the fitted regression model
                X_eval_te = np.column_stack([t_new[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                beta_est[te_ind,i] = mu_fit.predict(X_eval_te)
                    
        theta_est = np.diff(np.mean(beta_est, axis=0))/np.diff(t_new)
    return theta_est


def IPWDRDeriv(Y, X, t_eval, condTS_type, condTS_mod, L, h, kern="epanechnikov", 
               tau=0.001, b=None, self_norm=True):
    '''
    Estimating the derivative of a dose-response curve through the inverse 
    probability weighting (IPW) form under the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        condTS_type: str
            Specifying the model type for estimating the conditional density of
            the treatment variable T given the covariate vector S.
            
        condTS_mod: cikit-learn model or any python model that can use ".fit()" and ".predict()"
            The regression model for estimating the conditional density of T given S.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        h: float
            The bandwidth parameter.
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        tau: float
            The threshold value that lower bounds the estimated conditional density
            values. (Default: tau=0.001.)
            
        b: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
    '''
    kern_type = kern
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the conditional density model on the entire data
        if condTS_type == 'true':
            condTS_est = condTS_mod
        elif condTS_type == 'kde':
            condTS_est = CondDenEstKDE(X[:,0], X[:,1:], reg_mod=condTS_mod, y_eval=X[:,0], 
                                       x_eval=X[:,1:], kern=kern_type, b=b)
        else:
            condTS_est = CondDenEst(X[:,0], X[:,1:], reg_mod=condTS_mod, y_eval=X[:,0], 
                                    x_eval=X[:,1:], kern='gaussian', b=b)
        condTS_est[condTS_est < tau] = tau
        beta_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for i in range(t_eval.shape[0]):
            # Self-normalizing weights
            norm_w[i] = np.sum(kern((t_eval[i] - X[:,0])/h) / condTS_est) / h
            beta_hat[:,i] = ((X[:,0] - t_eval[i])/h) * kern((t_eval[i] - X[:,0])/h) * Y / (h**2 * sigmaK_sq * condTS_est)

        if self_norm:
            theta_est = np.sum(beta_hat, axis=0) / norm_w
        else:
            theta_est = np.mean(beta_hat, axis=0)
    else:
        # Conduct L-fold cross-fittings: fit the conditional density model on the training fold 
        # data and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        beta_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            X_te = X[te_ind,:]
            Y_te = Y[te_ind]
            
            if condTS_type == 'true':
                condTS_est = condTS_mod[te_ind]
            elif condTS_type == 'kde':
                condTS_est = CondDenEstKDE(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                           y_eval=X_te[:,0], x_eval=X_te[:,1:], kern=kern_type, b=b)
            else:
                condTS_est = CondDenEst(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                        y_eval=X_te[:,0], x_eval=X_te[:,1:], kern='gaussian', b=b)
            condTS_est[condTS_est < tau] = tau
            for i in range(t_eval.shape[0]):
                # Self-normalizing weights
                w = np.sum(kern((t_eval[i] - X[te_ind,0])/h) / condTS_est) / h
                if ~np.isnan(w) and w != np.inf:
                    norm_w[i] = norm_w[i] + w
                beta_hat[te_ind,i] = ((X[te_ind,0] - t_eval[i])/h) * kern((t_eval[i] - X[te_ind,0])/h) * Y_te / (h**2 * sigmaK_sq * condTS_est)

        if self_norm:
            norm_w[norm_w == 0] = 1
            theta_est = np.sum(beta_hat, axis=0) / norm_w
        else:
            theta_est = np.mean(beta_hat, axis=0)
    return theta_est


def DRDRDeriv(Y, X, t_eval, mu, condTS_type, condTS_mod, L, h, kern="epanechnikov", 
              n_iter=1000, lr=0.01, tau=0.001, b=None, self_norm=True):
    '''
    Estimating the derivative of a dose-response curve through the doubly robust
    (DR) form by a PyTorch neural network model under the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        mu: a neural network class defined by PyTorch
            The conditional mean outcome (or regression) model of Y given X.
            
        condTS_type: str
            Specifying the model type for estimating the conditional density of
            the treatment variable T given the covariate vector S.
            
        condTS_mod: cikit-learn model or any python model that can use ".fit()" and ".predict()"
            The regression model for estimating the conditional density of T given S.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        h: float
            The bandwidth parameter.
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        n_iter: int
            The number of iterations or training epochs of the neural network model.
            (Default: n_iter=1000.)
            
        lr: float
            The learning rate (Default: lr=0.01.)
            
        tau: float
            The threshold value that lower bounds the estimated conditional density
            values. (Default: tau=0.001.)
            
        b: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
            
        sd_est: (m,)-array
            The estimated asymptotic stdndard deviation of the DR derivative 
            estimator evaluated at points "t_eval".
    '''
    kern_type = kern
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the conditional density model and the regression model on the entire data
        if condTS_type == 'true':
            condTS_est = condTS_mod
        elif condTS_type == 'kde':
            condTS_est = CondDenEstKDE(X[:,0], X[:,1:], reg_mod=condTS_mod, 
                                       y_eval=X[:,0], x_eval=X[:,1:], kern=kern_type, b=b)
        else:
            condTS_est = CondDenEst(X[:,0], X[:,1:], reg_mod=condTS_mod, 
                                    y_eval=X[:,0], x_eval=X[:,1:], kern='gaussian', b=b)
        condTS_est[condTS_est < tau] = tau
        
        X_tensor = torch.from_numpy(X)
        Y_tensor = torch.from_numpy(Y.reshape(-1,1))
        NN_fit = train(mu, X_tensor, Y_tensor, lr=lr, n_epochs=n_iter)
        mu_hat = np.zeros((n, t_eval.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        beta_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the fitted regression model
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            X_eval_tensor = torch.from_numpy(X_eval)
            for j in range(X_eval.shape[0]):
                # Compute the gradient of the fitted regression model with respect to the first coordinate
                x = X_eval_tensor[j,:]
                x = x.clone().detach().requires_grad_(True)
                y = NN_fit(x)
                y_scalar = y[0]
                y_scalar.backward()
                x_grad = x.grad
                beta_hat[j,i] = x_grad[0].item()
            NN_fit.eval()
            mu_pred = NN_fit(X_eval_tensor)
            mu_hat[:,i] = mu_pred.detach().numpy()[:,0]
            
            IPW_hat[:,i] = ((X[:,0] - t_eval[i])/h) * kern((t_eval[i] - X[:,0])/h) * (Y - mu_hat[:,i] - (X[:,0] - t_eval[i])*beta_hat[:,i]) / (h**2 * sigmaK_sq * condTS_est)
            # Self-normalizing weights
            norm_w[i] = np.sum(kern((t_eval[i] - X[:,0])/h) / condTS_est) / (n * h)
        
        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        theta_hat = IPW_hat + beta_hat
        theta_est = np.mean(theta_hat, axis=0, where=~np.isnan(theta_hat))
        
        # Estimate the variance of theta(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (beta_hat[:,i] - theta_est[i]))**2 * (h**3)
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*(h**3)))
    else:
        # Conduct L-fold cross-fittings: fit the reciprocal of the conditional model 
        # and the regression model on the training fold data and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        mu_hat = np.zeros((n, t_eval.shape[0]))
        beta_hat = np.zeros((n, t_eval.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        cond_est_full = np.zeros((n,))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            Y_te = Y[te_ind]
            
            if condTS_type == 'true':
                condTS_est = condTS_mod[te_ind]
            elif condTS_type == 'kde':
                condTS_est = CondDenEstKDE(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                           y_eval=X_te[:,0], x_eval=X_te[:,1:], kern=kern_type, b=b)
            else:
                condTS_est = CondDenEst(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                        y_eval=X_te[:,0], x_eval=X_te[:,1:], kern='gaussian', b=b)
            condTS_est[condTS_est < tau] = tau
            cond_est_full[te_ind] = condTS_est
            
            X_tr_tensor = torch.from_numpy(X_tr)
            Y_tr_tensor = torch.from_numpy(Y_tr.reshape(-1,1))
            NN_fit = train(mu, X_tr_tensor, Y_tr_tensor, lr=lr, n_epochs=n_iter)
            for i in range(t_eval.shape[0]):
                # Define the data matrix for evaluating the fitted regression model
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                X_eval_te_tensor = torch.from_numpy(X_eval_te)
                for j in range(X_eval_te.shape[0]):
                    # Compute the gradient of the fitted regression model with respect to the first coordinate
                    x = X_eval_te_tensor[j,:]
                    x = x.clone().detach().requires_grad_(True)
                    y = NN_fit(x)
                    y_scalar = y[0]
                    y_scalar.backward()
                    x_grad = x.grad
                    beta_hat[j,i] = x_grad[0].item()
                    
                NN_fit.eval()
                mu_pred = NN_fit(X_eval_te_tensor)
                mu_hat[te_ind,i] = mu_pred.detach().numpy()[:,0]
                
                IPW_hat[te_ind,i] = ((X[te_ind,0] - t_eval[i])/h) * kern((t_eval[i] - X[te_ind,0])/h) * (Y_te - mu_hat[te_ind,i] - (X[te_ind,0] - t_eval[i])*beta_hat[te_ind,i]) / (h**2 * sigmaK_sq * condTS_est) 
                
                # Self-normalizing weights
                w = np.sum(kern((t_eval[i] - X[te_ind,0])/h) / condTS_est) / (n * h)
                norm_w[i] = norm_w[i] + w
        
        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        theta_hat = IPW_hat + beta_hat
        theta_est = np.mean(theta_hat, axis=0, where=~np.isnan(theta_hat))
        
        # Estimate the variance of theta(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (beta_hat[:,i] - theta_est[i]))**2 * (h**3)
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*(h**3)))
    return theta_est, sd_est



def DRDRDerivSKLearn(Y, X, t_eval, mu, condTS_type, condTS_mod, L, h, kern="epanechnikov", 
                     tau=0.001, b=None, delta=0.01, self_norm=True):
    '''
    Estimating the derivative of a dose-response curve through the doubly robust
    (DR) form under the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        mu: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        condTS_type: str
            Specifying the model type for estimating the conditional density of
            the treatment variable T given the covariate vector S.
            
        condTS_mod: cikit-learn model or any python model that can use ".fit()" and ".predict()"
            The regression model for estimating the conditional density of T given S.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        h: float
            The bandwidth parameter.
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        n_iter: int
            The number of iterations or training epochs of the neural network model.
            (Default: n_iter=1000.)
            
        lr: float
            The learning rate (Default: lr=0.01.)
            
        tau: float
            The threshold value that lower bounds the estimated conditional density
            values. (Default: tau=0.001.)
            
        b: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
        
        delta: float
            The step value for computing the finite differences (or numerical partial
            differentiation) of the fitted regression model.
        
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
            
        sd_est: (m,)-array
            The estimated asymptotic stdndard deviation of the DR derivative 
            estimator evaluated at points "t_eval".
    '''
    kern_type = kern
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the conditional density model and the regression model on the entire data
        if condTS_type == 'true':
            condTS_est = condTS_mod
        elif condTS_type == 'kde':
            condTS_est = CondDenEstKDE(X[:,0], X[:,1:], reg_mod=condTS_mod, 
                                       y_eval=X[:,0], x_eval=X[:,1:], kern=kern_type, b=b)
        else:
            condTS_est = CondDenEst(X[:,0], X[:,1:], reg_mod=condTS_mod, 
                                    y_eval=X[:,0], x_eval=X[:,1:], kern='gaussian', b=b)
        condTS_est[condTS_est < tau] = tau
        
        mu_fit = mu.fit(X, Y)
        
        mu_hat = np.zeros((n, t_eval.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        t_new = np.linspace(np.min(t_eval)-delta, np.max(t_eval)+delta, t_eval.shape[0]+1)
        beta_hat = np.zeros((n, t_new.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for i in range(t_new.shape[0]):
            X_new = np.column_stack([t_new[i]*np.ones(n), X[:,1:]])
            beta_hat[:,i] = mu_fit.predict(X_new)
        beta_hat = np.diff(beta_hat, axis=1)
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the fitted regression model
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            mu_hat[:,i] = mu_fit.predict(X_eval)
            
            IPW_hat[:,i] = ((X[:,0] - t_eval[i])/h) * kern((t_eval[i] - X[:,0])/h) * (Y - mu_hat[:,i] - (X[:,0] - t_eval[i])*beta_hat[:,i]) / (h**2 * sigmaK_sq * condTS_est)
            # Self-normalizing weights
            norm_w[i] = np.sum(kern((t_eval[i] - X[:,0])/h) / condTS_est) / (n * h)

        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        theta_hat = IPW_hat + beta_hat
        theta_est = np.mean(theta_hat, axis=0, where=~np.isnan(theta_hat))
        
        # Estimate the variance of theta(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (beta_hat[:,i] - theta_est[i]))**2 * (h**3)
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*(h**3)))
    else:
        # Conduct L-fold cross-fittings: fit the reciprocal of the conditional model 
        # and the regression model on the training fold data and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        mu_hat = np.zeros((n, t_eval.shape[0]))
        beta_hat = np.zeros((n, t_eval.shape[0]))
        
        t_new = np.linspace(np.min(t_eval)-delta, np.max(t_eval)+delta, t_eval.shape[0]+1)
        beta_can = np.zeros((n, t_new.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        cond_est_full = np.zeros((n,))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            Y_te = Y[te_ind]
            
            if condTS_type == 'true':
                condTS_est = condTS_mod[te_ind]
            elif condTS_type == 'kde':
                condTS_est = CondDenEstKDE(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                           y_eval=X_te[:,0], x_eval=X_te[:,1:], kern=kern_type, b=b)
            else:
                condTS_est = CondDenEst(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                        y_eval=X_te[:,0], x_eval=X_te[:,1:], kern='gaussian', b=b)
            condTS_est[condTS_est < tau] = tau
            cond_est_full[te_ind] = condTS_est
            
            mu_fit = mu.fit(X_tr, Y_tr)
            for i in range(t_new.shape[0]):
                X_new_te = np.column_stack([t_new[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                beta_can[te_ind,i] = mu_fit.predict(X_new_te)
            beta_hat[te_ind,:] = np.diff(beta_can[te_ind,:], axis=1)
            
            for i in range(t_eval.shape[0]):
                # Define the data matrix for evaluating the fitted regression model
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                mu_hat[te_ind,i] = mu_fit.predict(X_eval_te)
                
                IPW_hat[te_ind,i] = ((X[te_ind,0] - t_eval[i])/h) * kern((t_eval[i] - X[te_ind,0])/h) * (Y_te - mu_hat[te_ind,i] - (X[te_ind,0] - t_eval[i])*beta_hat[te_ind,i]) / (h**2 * sigmaK_sq * condTS_est) 
                    
                # Self-normalizing weights
                w = np.sum(kern((t_eval[i] - X[te_ind,0])/h) / condTS_est) / (n * h)
                norm_w[i] = norm_w[i] + w

        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        theta_hat = IPW_hat + beta_hat
        theta_est = np.mean(theta_hat, axis=0, where=~np.isnan(theta_hat))
        
        # Estimate the variance of theta(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (beta_hat[:,i] - theta_est[i]))**2 * (h**3)
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*(h**3)))
    return theta_est, sd_est


def DRDerivCurve(Y, X, t_eval=None, est="RA", beta_mod=None, n_iter=1000, lr=0.01, 
                 condTS_type=None, condTS_mod=None, L=1, h=None, kern="epanechnikov", 
                 tau=0.001, h_cond=None, delta=0.01, self_norm=True, print_bw=True):
    '''
    Dose-response curve derivative estimation under the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points. (Default: t_eval=None. 
            Then, t_eval=X[:,0], which consists of the observed treatment variables.)
            
        est: str
            The type of the dose-response curve estimator. (Default: est="RA". 
            Other choices include "IPW" and "DR".)
            
        beta_mod: PyTorch neural network class or scikit-learn model or any python 
        model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        n_iter: int
            The number of iterations or training epochs of the neural network model.
            (Default: n_iter=1000.)
            
        lr: float
            The learning rate (Default: lr=0.01.)
            
        condTS_type: str
            Specifying the model type for estimating the conditional density of
            the treatment variable T given the covariate vector S.
            
        condTS_mod: cikit-learn model or any python model that can use ".fit()" and ".predict()"
            The regression model for estimating the conditional density of T given S.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        h: float
            The bandwidth parameter for the IPW/DR estimator. (Default: h=None. 
            Then the Silverman's rule of thumb is applied; see Chen et al.(2016) 
            for details.)
            
        tau: float
            The threshold value that lower bounds the estimated conditional density
            values. (Default: tau=0.001.)
            
        h_cond: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
    
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
            
        sd_est: (m,)-array (if est="DR")
            The estimated asymptotic stdndard deviation of the DR derivative 
            estimator evaluated at points "t_eval".
    '''
    if t_eval is None: 
        t_eval = X[:,0].copy()
    
    n = X.shape[0]  ## Number of data points
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        h = (4/3)**(1/5)*(n**(-1/5))*np.std(X[:,0])
    
    if print_bw:
        print("The current bandwidth for the "+str(est)+" estimator is "+ str(h) + ".\n")
    
    if est == "RA":
        if isinstance(beta_mod, BaseEstimator):
            theta_est = RADRDerivSKLearn(Y, X, t_eval, beta_mod, L, delta)
        else:
            theta_est = RADRDeriv(Y, X, t_eval, beta_mod, L, n_iter, lr)
    elif est == "IPW":
        theta_est = IPWDRDeriv(Y, X, t_eval, condTS_type, condTS_mod, L, h, kern, 
                               tau, h_cond, self_norm)
    elif isinstance(beta_mod, BaseEstimator):
        theta_est = DRDRDerivSKLearn(Y, X, t_eval, beta_mod, condTS_type, condTS_mod, 
                                     L, h, kern, tau, h_cond, delta, self_norm)
    else:
        theta_est = DRDRDeriv(Y, X, t_eval, beta_mod, condTS_type, condTS_mod, L, 
                              h, kern, n_iter, lr, tau, h_cond, self_norm)
    
    return theta_est




#=======================================================================================#
# Implementations of the proposed estimators without assuming the positivity 
# condition (work for additive confounding models)

def RADRDerivBC(Y, X, t_eval, mu, L=1, n_iter=1000, lr=0.01, h_bar=None, 
                kernT_bar="gaussian", print_bw=False):
    '''
    Estimating the derivative of a dose-response curve through the regression 
    adjustment (or G-computation) form by a PyTorch neural network model without 
    assuming the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        mu: a neural network class defined by PyTorch
            The conditional mean outcome (or regression) model of Y given X.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        n_iter: int
            The number of iterations or training epochs of the neural network model.
            (Default: n_iter=1000.)
            
        lr: float
            The learning rate (Default: lr=0.01.)
            
        h_bar: float
            The bandwidth parameters for the Nadaraya-Watson conditional CDF estimator. 
            (Default: h_bar=None. Then, the Silverman's rule of thumb is applied. 
            See Chen et al.(2016) for details.)
            
        kernT_bar: str
            The name of the kernel function for Nadaraya-Watson conditional CDF 
            estimator. (Default: "gaussian".)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=False.)
            
    Return
    ----------
        theta_C: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
    '''
    n = X.shape[0]  ## Number of data points
    if h_bar is None:
        # Apply the Silverman's rule of thumb bandwidth in Chen et al. (2016)
        h_bar = (4/3)**(1/5)*(n**(-1/5))*np.std(X[:,0])
    if print_bw:
        print("The current bandwidth for the conditional CDF estimator is "+ str(h_bar) + ".\n")
    
    # Compute the weight matrix for NW conditional CDF estimator
    kernT_bar, sigmaK_sq, K_sq = KernelRetrieval(kernT_bar)
    weight_mat = kernT_bar((t_eval - X[:,0].reshape(-1,1)) / h_bar)
    weight_mat = weight_mat / np.sum(weight_mat, axis=0)
    weight_mat[np.isnan(weight_mat)] = 0
    
    if L <= 1:
        # No cross-fittings: fit the regression model on the entire data
        X_tensor = torch.from_numpy(X)
        Y_tensor = torch.from_numpy(Y.reshape(-1,1))
        NN_fit = train(mu, X_tensor, Y_tensor, lr=lr, n_epochs=n_iter)
        beta_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the fitted regression model
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            X_eval_tensor = torch.from_numpy(X_eval)
            for j in range(X_eval.shape[0]):
                # Compute the gradient of the fitted regression model with respect to the first coordinate
                x = X_eval_tensor[j,:]
                x = x.clone().detach().requires_grad_(True)
                y = NN_fit(x)
                y_scalar = y[0]
                y_scalar.backward()
                x_grad = x.grad
                beta_est[j,i] = x_grad[0].item()
            
        theta_C = np.sum(beta_est * weight_mat, axis=0)
    else:
        # Conduct L-fold cross-fittings: fit the regression model on the training fold data
        # and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        beta_est = np.zeros((n, t_eval.shape[0]))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            
            X_tr_tensor = torch.from_numpy(X_tr)
            Y_tr_tensor = torch.from_numpy(Y_tr.reshape(-1,1))
            NN_fit = train(mu, X_tr_tensor, Y_tr_tensor, lr=lr, n_epochs=n_iter)
            for i in range(t_eval.shape[0]):
                # Define the data matrix for evaluating the fitted regression model
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                X_eval_te_tensor = torch.from_numpy(X_eval_te)
                beta_hat = np.zeros((X_eval_te.shape[0],))
                for j in range(X_eval_te.shape[0]):
                    # Compute the gradient of the fitted regression model with respect to the first coordinate
                    x = X_eval_te_tensor[j,:]
                    x = x.clone().detach().requires_grad_(True)
                    y = NN_fit(x)
                    y_scalar = y[0]
                    y_scalar.backward()
                    x_grad = x.grad
                    beta_hat[j] = x_grad[0].item()
                beta_est[te_ind,i] = beta_hat
                    
        theta_C = np.sum(beta_est * weight_mat, axis=0)
    return theta_C


def IPWDRDerivBC(Y, X, t_eval, L=1, h=None, kern="epanechnikov", b=None, 
                 thres_val=0.75, self_norm=True):
    '''
    Estimating the derivative of a dose-response curve through the inverse 
    probability weighting (IPW) form without assuming the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        h: float
            The bandwidth parameter. (Default: h=None. Then, the Silverman's 
            rule of thumb is applied; see Chen et al.(2016) for details.)
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        b: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        thres_val: float
            The threshold factor that is multiplied to the maximum conditional 
            density values of S given T evaluated at the sample points. (Default: 
            thres_val=0.75.)
        
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
    '''
    kern_type = kern
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    n = X.shape[0]  ## Number of data points
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter
        h = (4/3)**(1/5)*(n**(-1/5))*np.std(X[:,0])
        print("The current bandwidth is "+ str(h) + ".\n")
        
    if L <= 1:
        # No cross-fittings: fit the density models on the entire data
        kde_joint = KDE(X, data=X, kern='gaussian', h=b)
        
        beta_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        condST_full = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the interior densities
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            new_joint = KDE(X_eval, data=X, kern=kern_type, h=b)
            t_den = KDE(t_eval[i], data=X[:,0], kern=kern_type, h=b)[0]
            condST_den = new_joint / t_den
            condST_full[:,i] = condST_den
            # condST_den = condST_den * (condST_den >= np.quantile(condST_den, thres_val)) / (1 - thres_val)
            trunc_perc = np.mean(condST_den >= thres_val * np.max(condST_den))
            condST_den = condST_den * (condST_den >= thres_val * np.max(condST_den)) / (1 - trunc_perc)
            den_fac = condST_den / kde_joint
            
            # Self-normalizing weights
            norm_w[i] = np.sum(kern((t_eval[i] - X[:,0])/h) * den_fac) / h
            beta_hat[:,i] = ((X[:,0] - t_eval[i])/h) * kern((t_eval[i] - X[:,0])/h) * Y * den_fac/ (h**2 * sigmaK_sq)

        if self_norm:
            theta_est = np.sum(beta_hat, axis=0) / norm_w
        else:
            theta_est = np.mean(beta_hat, axis=0)
    else:
        # Conduct L-fold cross-fittings: fit the conditional density model on the training fold 
        # data and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        beta_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        condST_full = np.zeros((n, t_eval.shape[0]))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            X_te = X[te_ind,:]
            Y_te = Y[te_ind]
            
            kde_joint = KDE(X_te, data=X_tr, kern='gaussian', h=b)
            for i in range(t_eval.shape[0]):
                # Define the data matrix for evaluating the interior densities
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                new_joint = KDE(X_eval_te, data=X_tr, kern='gaussian', h=b)
                t_den = KDE(t_eval[i], data=X_tr[:,0], kern='gaussian', h=b)[0]
                condST_den = new_joint / t_den
                condST_full[te_ind,i] = condST_den
                # condST_den = condST_den * (condST_den >= np.quantile(condST_den, thres_val)) / (1 - thres_val)
                trunc_perc = np.mean(condST_den >= thres_val * np.max(condST_den))
                condST_den = condST_den * (condST_den >= thres_val * np.max(condST_den)) / (1 - trunc_perc)
                den_fac = condST_den / kde_joint
                # Self-normalizing weights
                w = np.sum(kern((t_eval[i] - X_te[:,0])/h) * den_fac) / h
                if ~np.isnan(w) and w != np.inf:
                    norm_w[i] = norm_w[i] + w
                beta_hat[te_ind,i] = ((X_te[:,0] - t_eval[i])/h) * kern((t_eval[i] - X_te[:,0])/h) * Y_te * den_fac / (h**2 * sigmaK_sq)

        if self_norm:
            norm_w[norm_w == 0] = 1
            theta_est = np.sum(beta_hat, axis=0) / norm_w
        else:
            theta_est = np.mean(beta_hat, axis=0)
    return theta_est, condST_full


def DRDRDerivBC(Y, X, t_eval, mu, L=1, h=None, kern="epanechnikov", n_iter=1000, 
                lr=0.01, b=None, thres_val=0.75, self_norm=True):
    '''
    Estimating the derivative of a dose-response curve through the doubly robust 
    (DR) form by a PyTorch neural network model without assuming the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
        
        mu: a neural network class defined by PyTorch
            The conditional mean outcome (or regression) model of Y given X.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        h: float
            The bandwidth parameter. (Default: h=None. Then, the Silverman's 
            rule of thumb is applied; see Chen et al.(2016) for details.)
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        n_iter: int
            The number of iterations or training epochs of the neural network model.
            (Default: n_iter=1000.)
            
        lr: float
            The learning rate (Default: lr=0.01.)
            
        b: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        thres_val: float
            The threshold factor that is multiplied to the maximum conditional 
            density values of S given T evaluated at the sample points. (Default: 
            thres_val=0.75.)
        
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
    Return
    ----------
        theta_est: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
            
        sd_est: (m,)-array
            The estimated asymptotic stdndard deviation of the DR derivative 
            estimator evaluated at points "t_eval".
    '''
    kern_type = kern
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    n = X.shape[0]  ## Number of data points
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter
        h = (4/3)**(1/5)*(n**(-1/5))*np.std(X[:,0])
        print("The current bandwidth is "+ str(h) + ".\n")
        
    if L <= 1:
        # No cross-fittings: fit the density and regression model on the entire data
        kde_joint = KDE(X, data=X, kern='gaussian', h=b)
        
        X_tensor = torch.from_numpy(X)
        Y_tensor = torch.from_numpy(Y.reshape(-1,1))
        NN_fit = train(mu, X_tensor, Y_tensor, lr=lr, n_epochs=n_iter)
        mu_hat = np.zeros((n, t_eval.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        beta_hat = np.zeros((n, t_eval.shape[0]))
        condST_int = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the interior densities and fitted regression model
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            new_joint = KDE(X_eval, data=X, kern=kern_type, h=b)
            t_den = KDE(t_eval[i], data=X[:,0], kern=kern_type, h=b)[0]
            condST_den = new_joint / t_den
            # condST_den = condST_den * (condST_den >= np.quantile(condST_den, thres_val)) / (1 - thres_val)
            trunc_perc = np.mean(condST_den >= thres_val * np.max(condST_den))
            condST_den = condST_den * (condST_den >= thres_val * np.max(condST_den)) / (1 - trunc_perc)
            den_fac = condST_den / kde_joint
            
            X_eval_tensor = torch.from_numpy(X_eval)
            for j in range(X_eval.shape[0]):
                # Compute the gradient of the fitted regression model with respect to the first coordinate
                x = X_eval_tensor[j,:]
                x = x.clone().detach().requires_grad_(True)
                y = NN_fit(x)
                y_scalar = y[0]
                y_scalar.backward()
                x_grad = x.grad
                beta_hat[j,i] = x_grad[0].item()
            NN_fit.eval()
            mu_pred = NN_fit(X_eval_tensor)
            mu_hat[:,i] = mu_pred.detach().numpy()[:,0]
            
            IPW_hat[:,i] = ((X[:,0] - t_eval[i])/h) * kern((t_eval[i] - X[:,0])/h) * (Y - mu_hat[:,i] - (X[:,0] - t_eval[i])*beta_hat[:,i]) * den_fac/ (h**2 * sigmaK_sq)
            
            # Self-normalizing weights
            norm_w[i] = np.sum(kern((t_eval[i] - X[:,0])/h) * den_fac) / (n * h)
            condST_int[:,i] = condST_den
            
        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        theta_hat = IPW_hat + beta_hat * condST_int
        theta_est = np.mean(theta_hat, axis=0, where=~np.isnan(theta_hat))
        
        # Estimate the variance of theta(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (np.mean(beta_hat[:,i] * condST_int[:,i]) - theta_est[i]))**2 * (h**3)
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*(h**3)))
    else:
        # Conduct L-fold cross-fittings: fit the reciprocal of the conditional model 
        # and the regression model on the training fold data and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        mu_hat = np.zeros((n, t_eval.shape[0]))
        beta_hat = np.zeros((n, t_eval.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        condST_int = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            Y_te = Y[te_ind]
            
            kde_joint = KDE(X_te, data=X_tr, kern='gaussian', h=b)
            
            X_tr_tensor = torch.from_numpy(X_tr)
            Y_tr_tensor = torch.from_numpy(Y_tr.reshape(-1,1))
            NN_fit = train(mu, X_tr_tensor, Y_tr_tensor, lr=lr, n_epochs=n_iter)
            for i in range(t_eval.shape[0]):
                # Define the data matrix for evaluating the interior densities and fitted regression model
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                new_joint = KDE(X_eval_te, data=X_tr, kern='gaussian', h=b)
                t_den = KDE(t_eval[i], data=X_tr[:,0], kern='gaussian', h=b)[0]
                condST_den = new_joint / t_den
                # condST_den = condST_den * (condST_den >= np.quantile(condST_den, thres_val)) / (1 - thres_val)
                trunc_perc = np.mean(condST_den >= thres_val * np.max(condST_den))
                condST_den = condST_den * (condST_den >= thres_val * np.max(condST_den)) / (1 - trunc_perc)
                den_fac = condST_den / kde_joint
                
                X_eval_te_tensor = torch.from_numpy(X_eval_te)
                for j in range(X_eval_te.shape[0]):
                    # Compute the gradient of the fitted regression model with respect to the first coordinate
                    x = X_eval_te_tensor[j,:]
                    x = x.clone().detach().requires_grad_(True)
                    y = NN_fit(x)
                    y_scalar = y[0]
                    y_scalar.backward()
                    x_grad = x.grad
                    beta_hat[j,i] = x_grad[0].item()
                    
                NN_fit.eval()
                mu_pred = NN_fit(X_eval_te_tensor)
                mu_hat[te_ind,i] = mu_pred.detach().numpy()[:,0]
                
                IPW_hat[te_ind,i] = ((X[te_ind,0] - t_eval[i])/h) * kern((t_eval[i] - X[te_ind,0])/h) * (Y_te - mu_hat[te_ind,i] - (X[te_ind,0] - t_eval[i])*beta_hat[te_ind,i]) * den_fac / (h**2 * sigmaK_sq) 
                # Self-normalizing weights
                w = np.sum(kern((t_eval[i] - X[te_ind,0])/h) * den_fac) / (n * h)
                norm_w[i] = norm_w[i] + w
                condST_int[te_ind,i] = condST_den
        
        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        theta_hat = IPW_hat + beta_hat * condST_int
        theta_est = np.mean(theta_hat, axis=0, where=~np.isnan(theta_hat))
        
        # Estimate the variance of theta(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (np.mean(beta_hat[:,i] * condST_int[:,i]) - theta_est[i]))**2 * (h**3)
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*(h**3)))
    return theta_est, sd_est




# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: May 19, 2024

# Description: This script contains the utility functions for the main functions 
# in our package.

import numpy as np
from .rbf import KernelRetrieval
import scipy.integrate as integrate
from sklearn.preprocessing import PolynomialFeatures

#=======================================================================================#


def RoTBWLocalPoly(Y, X, kernT="epanechnikov", kernS="epanechnikov", C_h=7, C_b=3):
    '''
    Compute the rule-of-thumb bandwidth selector in Eq.(A1) of 
    Yang and Tschernig (1999).
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
        
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
        C_h,C_b: float
            The scaling factors for the rule-of-thumb bandwidth parameters. 
            (Default: C_h=7, C_b=3.)
    
    Return
    ----------
        h: float
            The rule-of-thumb bandwidth parameter for the treatment/exposure variable.
            
        b: (d,)-array
            The rule-of-thumb bandwidth vector for the confounding variables.
    '''
    n = X.shape[0]  ## Number of data points
    d = X.shape[1] - 1
    
    kernT, sigmaK_sq, K_sq = KernelRetrieval(kernT)
    # Apply the rule-of-thumb bandwidth selector in Eq.(A1) of Yang and Tschernig (1999)
    p_coeff = np.polyfit(X[:,0], Y, 4)
    sec_deriv = 12*p_coeff[0]*X[:,0] + 6*p_coeff[1]*X[:,0] + 2*X[:,0]
    C_fun = np.mean(sec_deriv**2)
    T = X[:,0].reshape(-1,1)
    lhs = np.concatenate([np.ones((n,1)), X, T**2, T**3, T**4], axis=1)
    rcond = np.finfo(lhs.dtype).eps * max(*lhs.shape)
    beta = np.linalg.lstsq(lhs, Y, rcond=rcond)[0]
    # Compute the residual sum of squares
    resid = np.sum((Y - np.dot(lhs, beta))**2) * (np.max(X[:,0]) - np.min(X[:,0]))/ (n-5)
    sigmaK_sq = sigmaK_sq**2
    # ROT
    h = ((K_sq*resid)/(4*n*sigmaK_sq*C_fun))**(1/5) * (n**(d/(5*(d+5)))) * C_h
    
    kernS, sigmaK_sq, K_sq = KernelRetrieval(kernS)
    sec_deriv = np.zeros((n, d))
    for i in range(1, d+1):
        # Fit a fourth-order polynomial to each confounding variable
        p_coeff = np.polyfit(X[:,i], Y, 4)
        sec_deriv[:,i-1] = 12*p_coeff[0]*X[:,i] + 6*p_coeff[1]*X[:,i] + 2*X[:,i]
    C_fun = np.sum(np.diag(np.dot(sec_deriv.T, sec_deriv))/n)
    lhs = np.concatenate([np.ones((n,1)), X, T**2, T**3, T**4], axis=1)
    # lhs = np.concatenate([np.ones((n,1)), X[:,1:]], axis=1)
    rcond = np.finfo(lhs.dtype).eps * max(*lhs.shape)
    beta = np.linalg.lstsq(lhs, Y, rcond=rcond)[0]
    resid = np.sum((Y - np.dot(lhs, beta))**2) * (np.max(X[:,1:], axis=0) - np.min(X[:,1:], axis=0)) / (n-5)
    sigmaK_sq = sigmaK_sq**2
    K_sq = K_sq**d
    b = ((K_sq*d*resid)/(4*n*sigmaK_sq*C_fun))**(1/(d+5)) * C_b
    
    return h, b


def HatMatrix(X, degree=2, deriv_ord=1, h=None, b=None, print_bw=True, 
              kernT="epanechnikov", kernS="epanechnikov"):
    '''
    Compute the hat matrix of the local polynomial regression when it is viewed 
    as a linear smoother.
    
    Parameters
    ----------
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome function. 
            (Default: deriv_ord=1. Then, it estimates the partial derivative of the 
            conditional mean outcome function with respect to the treatment variable.)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. 
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
    Return
    ----------
        hat_mat: (n,n)-array
            The hat matrix.
    '''
    n = X.shape[0]  ## Number of data points
    d = X.shape[1] - 1
    x_eval = X.copy()
    
    if print_bw:
        print("The current bandwidth for treament variable in the local polynomial regression is "+ str(h) + ".\n")
        print("The current bandwidth for confounding variables in the local polynomial regression is "+ str(b) + ".\n")
    
    kernT, sigmaK_sq, K_sq = KernelRetrieval(kernT)
    kernS, sigmaK_sq, K_sq = KernelRetrieval(kernS)
    hat_mat = np.zeros((n, n))
    for i in range(x_eval.shape[0]):
        weights = kernT((X[:,0] - x_eval[i,0])/h) * np.prod(kernS((X[:,1:] - x_eval[i,1:])/b), axis=1)
        # Filter out the data points with zero weights to speed up regressions with kernels of compact support
        inds = np.where(np.abs(weights) > 1e-26)[0]
        X_dat = np.zeros((n, degree+1+d))
        # X_dat[:,:(degree+1)] = (((X[:,0] - x_eval[i,0])/h).reshape(-1,1))**(np.arange(degree+1))
        for p in range(degree + 1):
            X_dat[:,p] = ((X[:,0] - x_eval[i,0])/h)**p
        X_dat[:,(degree+1):] = (X[:,1:] - x_eval[i,1:])/b
        X_dat = X_dat[inds,:]
        W = np.diag(weights[inds])
        design_mat = np.dot(np.dot(X_dat.T, W), X_dat)
        try:
            hat_mat_samp = np.dot(np.linalg.inv(design_mat), np.dot(X_dat.T, W))
        except:
            # Add some small quantities to the diagonal matrix to prevent the singularity of the matrix
            design_mat = design_mat + 1e-16*np.eye(design_mat.shape[0])
            hat_mat_samp = np.dot(np.linalg.inv(design_mat), np.dot(X_dat.T, W))
        hat_mat[i,inds] = hat_mat_samp[deriv_ord,:]
    return hat_mat
    
    
    

def BndKern(x_qry, kern, deriv_ord=0, alpha=1, bnd='left'):
    '''
    Generalized jackknife boundary kernel.
    
    Parameters
    ----------
        x_qry: (m,)-array
            The coordinates of m query points in the 1-dimensional Euclidean space.
       
        kern: python function
            The kernel function.
            
        deriv_ord: int
            The order of the derivative estimator. (Default: deriv_ord=0, which
            is for nonparametric density or curve estimation.)
        
        alpha: float
            The truncated proportion of the kernel support (0 <= alpha <= 1). 
            (Default: alpha=1, which recovers the original kernel function for 
             the interior points.)
        
        bnd: str
            Indicator of whether the input point is within the left or right 
            boundary of the support. (Default: bnd='left'.)
    
    Return
    ----------
        res: (m,)-array
            The boundary kernel function evaluated at m query points.
    '''
    kappa_lst = np.zeros((2*deriv_ord+3,))
    if bnd == 'left':
        sign = -1
    else:
        sign = 1
    for i in range(deriv_ord, 3*deriv_ord + 3):
        kappa_lst[i-deriv_ord] = integrate.quad(lambda x: ((sign*x)**i)*kern(x), 
                                                -np.inf, alpha)[0]
    mat = np.zeros((deriv_ord + 2, deriv_ord + 2))
    for i in range(deriv_ord + 2):
        mat[i,:] = kappa_lst[i:(i+deriv_ord+2)]
    B = np.zeros((deriv_ord + 2, ))
    B[deriv_ord] = 1
    coef = np.linalg.solve(mat, B)
    res = 0
    for j in range(deriv_ord+2):
        res += coef[j]*(x_qry**j)*kern(x_qry)
    return res


def KDE1D(x, data, kern='epanechnikov', h=None):
    '''
    One-dimensional kernel density estimation with generalized jackknife boundary
    corrections (Jones 1993).
    
    Parameters
    ----------
        x: (m,)-array
            The coordinates of m query points in the 1-dim Euclidean space.
    
        data: (n,)-array
            The coordinates of n random sample points in the d-dimensional 
            Euclidean space.
       
        kern: str
            The name of the kernel function. (Default: "epanechnikov".)
            
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied; see Chen et al.(2016) for details.)
    
    Return
    ----------
        f_hat: (m,)-array
            The corresponding kernel density estimates at m query points.
    '''
    n = data.shape[0]  ## Number of data points
    d = 1  ## Dimension of the data
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        h = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.std(data)
        print("The current bandwidth for KDE is "+ str(h) + ".\n")
    
    f_hat = np.zeros((x.shape[0], ))
    for i in range(x.shape[0]):
        if x[i] < np.min(data) or x[i] > np.max(data):
            f_hat[i] = 0
        elif x[i] < np.min(data) + h:
            alpha = (x[i] - np.min(data))/h
            f_hat[i] = np.mean(BndKern((data - x[i])/h, kern=kern, deriv_ord=0, 
                                       alpha=alpha, bnd='left') / h)
        elif x[i] > np.max(data) - h:
            alpha = (np.max(data) - x[i])/h
            f_hat[i] = np.mean(BndKern((data - x[i])/h, kern=kern, deriv_ord=0, 
                                       alpha=alpha, bnd='right') / h)
        else:
            f_hat[i] = np.mean(kern((data - x[i])/h) / h)
    return f_hat


def KDE(x, data, kern='gaussian', h=None):
    '''
    The d-dimensional Euclidean kernel density estimator.
    
    Parameters
    ----------
        x: (m,d)-array
            The coordinates of m query points in the d-dim Euclidean space.
    
        data: (n,d)-array
            The coordinates of n random sample points in the d-dimensional 
            Euclidean space.
            
        kern: str
            The name of the kernel function. (Default: "gaussian".)
       
        h: float
            The bandwidth parameter. (Default: h=None. Then the Silverman's 
            rule of thumb is applied. See Chen et al.(2016) for details.)
    
    Return
    ----------
        f_hat: (m,)-array
            The corresponding kernel density estimates at m query points.
    '''
    n = data.shape[0]  ## Number of data points
    if len(data.shape) == 1:
        d = 1
        data = data.reshape(-1,1)
        x = x.reshape(-1,1)
    else:
        d = data.shape[1]  ## Dimension of the data
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        h = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.std(data, axis=0)
        print("The current bandwidth is "+ str(h) + ".\n")
    
    f_hat = np.zeros((x.shape[0], ))
    for i in range(x.shape[0]):
        f_hat[i] = np.mean(np.exp(np.sum(-((x[i,:] - data)/h)**2, axis=1)/2))/ ((2*np.pi)**(d/2)*np.prod(h))
    return f_hat


def CondDenEstKDE(Y, X, reg_mod, y_eval=None, x_eval=None, kern='epanechnikov', b=None):
    '''
    Conditional density estimation by applying the kernel density estimator (KDE) 
    on the regression residuals.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d)-array
            The d-dimensional covariates of n observations.
            
        reg_mod: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        y_eval: (m,)-array
            The outcome variables at which we evaluate the estimated conditional
            densities.
            
        x_eval: (m,d)-array
            The covariates at which we evaluate the estimated conditional
            densities.
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        b: float
            The bandwidth parameter for KDE. (Default: b=None.)
    
    Return
    ----------
        cond_est: (m,)-array
            The estimated conditional densities at the m query points.
    '''
    if y_eval is None:
        y_eval = Y
    if x_eval is None:
        x_eval = X
    
    reg_hat = reg_mod.fit(X, Y)
    y_pred = reg_hat.predict(x_eval)
    
    cond_est = KDE1D(x=y_eval-y_pred, data=y_eval-y_pred, kern=kern, h=b)
    return cond_est


def CondDenEst(Y, X, reg_mod, y_eval=None, x_eval=None, kern='gaussian', b=None, 
               poly_ext=False):
    '''
    Conditional density estimation via nonparametric regression on the kernel-smoothed 
    outcome variables.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d)-array
            The d-dimensional covariates of n observations.
            
        reg_mod: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        y_eval: (m,)-array
            The outcome variables on which we evaluate the estimated conditional
            densities.
            
        x_eval: (m,d)-array
            The covariates on which we evaluate the estimated conditional
            densities.
            
        kern: str
            The name of the kernel function. (Default: kern="gaussian".)
            
        b: float
            The bandwidth parameter for KDE. (Default: b=None.)
        
        poly_ext: boolean
            The indicator of whether polynomial features are generated from the 
            current covariates. (Default: poly_ext=False.)
    
    Return
    ----------
        cond_est: (m,)-array
            The estimated conditional densities at the m query points.
    '''
    n = X.shape[0]
    
    if b is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        b = (4/3)**(1/5)*(n**(-1/5))*np.std(Y)
        print("The current bandwidth for the conditional density estimator is "+ str(b) + ".\n")
    
    if y_eval is None:
        y_eval = Y
    if x_eval is None:
        x_eval = X
    
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    cond_est = np.zeros((y_eval.shape[0],))
    if poly_ext:
        X_inter = PolynomialFeatures(degree=3, interaction_only=False, 
                                     include_bias=True).fit_transform(X)
        x_eval_inter = PolynomialFeatures(degree=3, interaction_only=False, 
                                          include_bias=True).fit_transform(x_eval)
    else:
        X_inter = X
        x_eval_inter = x_eval
    
    # Estimate the conditional density by fitting the kernelized outcomes
    for i in range(y_eval.shape[0]):
        Y_kern = kern((Y - y_eval[i])/b)/b
        reg_hat = reg_mod.fit(X_inter, Y_kern)
        y_pred = reg_hat.predict(x_eval_inter[i,:].reshape(1,-1))
        cond_est[i] = y_pred[0]
    
    return cond_est




# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: March 19, 2024

# Description: This file contains the implementations of common kernel functions.

import numpy as np

#=======================================================================================#

def rectangular(t):
    '''
    Rectangular/uniform kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    ind = (np.abs(t)<=1)
    res = np.abs(0.5*ind)
    return res

def triangular(t):
    '''
    Triangular kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    ind = (np.abs(t)<=1)
    res = np.abs((1-np.abs(t))*ind)
    return res

def epanechnikov(t):
    '''
    Epanechnikov kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    ind = (np.abs(t)<=1)
    res = np.abs(0.75*(1-t**2)*ind)
    return res

def biweight(t):
    '''
    Biweight/quartic kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    ind = (np.abs(t)<=1)
    res = np.abs(((15/16)*(1-t**2)**2)*ind)
    return res

def triweight(t):
    '''
    Triweight kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    ind = (np.abs(t)<=1)
    res = np.abs((35/32)*(1-t**2)**3*ind)
    return res

def tricube(t):
    '''
    Tricube kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    ind = (np.abs(t)<=1)
    res = np.abs((70/81)*(1-np.abs(t)**3)**3*ind)
    return res

def gaussian(t):
    '''
    Gaussian kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    res = (1/np.sqrt(2*np.pi))*np.exp(-0.5*t**2)
    return res

def bigaussian(t):
    '''
    Bigaussian kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    res = (2/np.sqrt(np.pi))*(t**2)*np.exp(-t**2)
    return res

def cosine(t):
    '''
    Cosine kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    ind = (np.abs(t)<=1)
    res = np.abs((np.pi/4)*np.cos(np.pi*t/2)*ind)
    return res

def logistic(t):
    '''
    Logistic kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    res = 1/(np.exp(t)+2+np.exp(-t))
    return res

def sigmoid(t):
    '''
    Sigmoid kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    res = (2/np.pi)/(np.exp(t)+np.exp(-t))
    return res

def silverman(t):
    '''
    Silverman kernel function.

    Parameters
    ----------
        t: float or (n,)-array
            The query points.

    Return
    -------
        res: float or (n,)-array
            The kernel values evaluated at the query points.
    '''
    res = 0.5*np.exp(-np.abs(t)/np.sqrt(2))*np.sin(np.abs(t)/np.sqrt(2)+np.pi/4)
    return res


def KernelRetrieval(name):
    '''
    Retrieving the kernel function, its second moment, and its variance based on 
    the name.
    
    Parameters
    ----------
        name: str
            The name of the kernel function.

    Return
    --------
        kern_func: python function
            The kernel function.
            
        sigmaK_sq: float
            The second moment of the kernel function.
            
        K_sq: float
            The variance of the kernel function.
    '''
    if name == "rectangular":
        return rectangular, 1/3, 1/2
    
    if name == "triangular":
        return triangular, 1/6, 2/3
    
    if name == "epanechnikov":
        return epanechnikov, 1/5, 3/5
    
    if name == "biweight":
        return biweight, 1/7, 5/7
    
    if name == "triweight":
        return triweight, 1/9, 350/429
    
    if name == "tricube":
        return triangular, 35/243, 175/247
    
    if name == "gaussian":
        return gaussian, 1, 1/(2*np.sqrt(np.pi))
    
    if name == "bigaussian":
        return bigaussian, 3/2, 3*np.sqrt(2/np.pi)/8
    
    if name == "cosine":
        return cosine, 1-8/(np.pi**2), np.pi**2/16
    
    if name == "logistic":
        return logistic, np.pi**2/3, 1/6
    
    if name == "sigmoid":
        return sigmoid, np.pi**2/4, 2/(np.pi**2)
    
    if name == "silverman":
        return silverman, 0, 3*np.sqrt(2)/16


"""Nonparametric Inference on Dose-Response Curve and its Derivative: With and Without Positivity"""

from .npDoseResponse import IntegEst, DerivEffect, IntegEstBoot, DerivEffectBoot, LocalPolyReg, LocalPolyReg1D, RegAdjust
from .npDoseResponseDR import RegAdjustDR, IPWDR, DRDR, DRCurve
from .npDoseResponseDerivDR import NeurNet, train, RADRDeriv, RADRDerivSKLearn, IPWDRDeriv, DRDRDeriv, DRDRDerivSKLearn, DRDerivCurve, RADRDerivBC, IPWDRDerivBC, DRDRDerivBC

__author__ = "Yikun Zhang"
__version__ = "0.0.10"



# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: May 23, 2024

# Description: This script contains the implementation of local polynomial regression 
# as well as our proposed integral estimator and its localized derivative estimator 
# (bias-corrected RA estimator) without the positivity condition.

import numpy as np
from .rbf import KernelRetrieval
from multiprocessing import Pool
from functools import partial
from .utils import RoTBWLocalPoly, HatMatrix

#=======================================================================================#

def LocalPolyReg(Y, X, x_eval=None, degree=2, deriv_ord=1, h=None, b=None, C_h=7, 
                 C_b=3, print_bw=True, kernT="epanechnikov", kernS="epanechnikov", 
                 h_lst=np.linspace(0.5, 15, 30), b_lst=np.linspace(0.2, 6, 30)):
    '''
    (Partial) Local polynomial regression for estimating the conditional mean outcome 
    function and its partial derivatives. We use higher order local monomials for 
    the treatment variable and first-order local monomials for the confounding variables.
    
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
        
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        x_eval: (m,d+1)-array
            The coordinates of the m evaluation points. (Default: x_eval=None. 
            Then, x_eval=X.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome 
            function. (Default: deriv_ord=1. Then, it estimates the partial 
            derivative of the conditional mean outcome function with respect to 
            the treatment variable.)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. (Default: h=None, b=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used 
            with additional scaling factors C_h and C_b, respectively.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
        h_lst, b_lst: (k1,)-array and (k2,)-array
            Candidate searching values of h,b for LOOCV.
            
    Return
    ----------
        Y_est: (m,)-array
            The estimated conditional mean outcome function or its partial derivatives 
            evaluated at points "x_eval".
    '''
    if x_eval is None: 
        x_eval = X.copy()
        
    if (h is None) and (b is None):
        # Apply the rule-of-thumb bandwidth selector in Eq.(A1) of Yang and Tschernig (1999)
        h, b = RoTBWLocalPoly(Y, X, kernT=kernT, kernS=kernS, C_h=C_h, C_b=C_b)
    elif h is None:
        h, b_can = RoTBWLocalPoly(Y, X, kernT=kernT, kernS=kernS, C_h=C_h, C_b=C_b)
    elif b is None:
        h_can, b = RoTBWLocalPoly(Y, X, kernT=kernT, kernS=kernS, C_h=C_h, C_b=C_b)
        
    if (h == "cv") and (b == "cv"):
        cv_mse = np.zeros((len(h_lst), len(b_lst)))
        for i in range(len(h_lst)):
            for j in range(len(b_lst)):
                Y_reg_est = LocalPolyRegMain(Y, X, x_eval=X, degree=1, deriv_ord=0, 
                                             h=h_lst[i], b=b_lst[i], 
                                             kernT=kernT, kernS=kernS)
                # Leave-one-out CV (with hat matrix trick)
                hat_mat = HatMatrix(X, degree=2, deriv_ord=1, h=h_lst[i], b=b_lst[i], kernT=kernT, kernS=kernS)
                cv_mse[i,j] = np.mean(((Y - Y_reg_est) / (1 - np.diag(hat_mat)))**2)
        argmin_ind = np.unravel_index(cv_mse.argmin(), cv_mse.shape)
        h_opt = h_lst[argmin_ind[0]]
        b_opt = b_lst[argmin_ind[1]]
        h = h_opt
        b = b_opt
    
    if print_bw:
        print("The current bandwidth for treament variable in the local polynomial regression is "+ str(h) + ".\n")
        print("The current bandwidth for confounding variables in the local polynomial regression is "+ str(b) + ".\n")
    
    Y_est = LocalPolyRegMain(Y, X, x_eval=x_eval, degree=degree, deriv_ord=deriv_ord, 
                             h=h, b=b, kernT=kernT, kernS=kernS)
    return Y_est


def LocalPolyRegMain(Y, X, x_eval=None, degree=2, deriv_ord=1, h=None, b=None, 
                     kernT="epanechnikov", kernS="epanechnikov"):
    '''
    Main function for computing the local polynomial regression.
    
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
        
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        x_eval: (m,d+1)-array
            The coordinates of the m evaluation points. (Default: x_eval=None. 
            Then, x_eval=X.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome 
            function. (Default: deriv_ord=1. Then, it estimates the partial 
            derivative of the conditional mean outcome function with respect to 
            the treatment variable.)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables.
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
    Return
    ----------
        Y_est: (m,)-array
            The estimated conditional mean outcome function or its partial derivatives 
            evaluated at points "x_eval".
    '''
    kernT, sigmaK_sq, K_sq = KernelRetrieval(kernT)
    kernS, sigmaK_sq, K_sq = KernelRetrieval(kernS)
    
    n = X.shape[0]  ## Number of data points
    d = X.shape[1] - 1
    Y_est = np.zeros((x_eval.shape[0],))
    for i in range(x_eval.shape[0]):
        weights = kernT((X[:,0] - x_eval[i,0])/h) * np.prod(kernS((X[:,1:] - x_eval[i,1:])/b), axis=1)
        # Filter out the data points with zero weights to speed up regressions with kernels of compact support
        inds = np.where(np.abs(weights) > 1e-26)[0]
        X_dat = np.zeros((n, degree+1+d))
        # X_dat[:,:(degree+1)] = ((X[:,0] - x_eval[i,0]).reshape(-1,1))**(np.arange(degree+1))
        for p in range(degree + 1):
            X_dat[:,p] = (X[:,0] - x_eval[i,0])**p
        X_dat[:,(degree+1):] = X[:,1:] - x_eval[i,1:]
        weight_sqrt = np.sqrt(weights)[inds]
        lhs = np.dot(np.diag(weight_sqrt), X_dat[inds,:])
        rhs = weight_sqrt*Y[inds]
        rcond = np.finfo(lhs.dtype).eps * max(*lhs.shape)
        beta = np.linalg.lstsq(lhs, rhs, rcond=rcond)[0]
        Y_est[i] = np.math.factorial(deriv_ord)*beta[deriv_ord]
    return Y_est


def LocalPolyReg_Fs(x_eval, Y, X, degree=2, deriv_ord=1, h=None, b=None, C_h=7, 
                    C_b=3, print_bw=True, kernT="epanechnikov", kernS="epanechnikov", 
                    h_lst=np.linspace(0.5, 15, 30), b_lst=np.linspace(0.2, 6, 30)):
    '''
    (Partial) Local polynomial regression for estimating the conditional mean outcome 
    function and its partial derivatives. We use higher order local monomials for 
    the treatment variable and first-order local monomials for the confounding variables.
    (This function is for multi-process execution only.)
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
        
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        x_eval: (m,d+1)-array
            The coordinates of the m evaluation points. (Default: x_eval=None. 
            Then, x_eval=X.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome 
            function. (Default: deriv_ord=1. Then, it estimates the partial 
            derivative of the conditional mean outcome function with respect to 
            the treatment variable.)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. (Default: h=None, b=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used
            with additional scaling factors C_h and C_b, respectively.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
        h_lst, b_lst: (k1,)-array and (k2,)-array
            Candidate searching values of h,b for LOOCV.
            
    Return
    ----------
        Y_est: (m,)-array
            The estimated conditional mean outcome function or its partial derivatives 
            evaluated at points "x_eval".
    '''
    if x_eval is None: 
        x_eval = X.copy()
        
    if (h is None) and (b is None):
        # Apply the rule-of-thumb bandwidth selector in Eq.(A1) of Yang and Tschernig (1999)
        h, b = RoTBWLocalPoly(Y, X, kernT=kernT, kernS=kernS, C_h=C_h, C_b=C_b)
    elif h is None:
        h, b_can = RoTBWLocalPoly(Y, X, kernT=kernT, kernS=kernS, C_h=C_h, C_b=C_b)
    elif b is None:
        h_can, b = RoTBWLocalPoly(Y, X, kernT=kernT, kernS=kernS, C_h=C_h, C_b=C_b)
        
    if (h == "cv") and (b == "cv"):
        cv_mse = np.zeros((len(h_lst), len(b_lst)))
        for i in range(len(h_lst)):
            for j in range(len(b_lst)):
                Y_reg_est = LocalPolyRegMain(Y, X, x_eval=X, degree=1, deriv_ord=0, 
                                             h=h_lst[i], b=b_lst[i], 
                                             kernT=kernT, kernS=kernS)
                # Leave-one-out CV (with hat matrix trick)
                hat_mat = HatMatrix(X, degree=2, deriv_ord=1, h=h_lst[i], b=b_lst[i], kernT=kernT, kernS=kernS)
                cv_mse[i,j] = np.mean(((Y - Y_reg_est) / (1 - np.diag(hat_mat)))**2)
        argmin_ind = np.unravel_index(cv_mse.argmin(), cv_mse.shape)
        h_opt = h_lst[argmin_ind[0]]
        b_opt = b_lst[argmin_ind[1]]
        h = h_opt
        b = b_opt
    
    if print_bw:
        print("The current bandwidth for treament variable in the local polynomial regression is "+ str(h) + ".\n")
        print("The current bandwidth for confounding variables in the local polynomial regression is "+ str(b) + ".\n")
    
    Y_est = LocalPolyRegMain(Y, X, x_eval=x_eval, degree=degree, deriv_ord=deriv_ord, 
                             h=h, b=b, kernT=kernT, kernS=kernS)
    return Y_est


def DerivEffect(Y, X, t_eval=None, h_bar=None, kernT_bar="gaussian", h=None, b=None, 
                C_h=7, C_b=3, print_bw=True, degree=2, deriv_ord=1, kernT="epanechnikov", 
                kernS="epanechnikov", parallel=False, processes=20):
    '''
    Estimating the derivative of the dose-response curve via Nadaraya-Watson 
    conditional CDF estimator.
    
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points. (Default: t_eval=None. 
            Then, t_eval=X[:,0], which consists of the observed treatment variables.)
            
        h_bar: float
            The bandwidth parameters for the Nadaraya-Watson conditional CDF estimator. 
            (Default: h_bar=None. Then, the Silverman's rule of thumb is applied. 
            See Chen et al.(2016) for details.)
            
        kernT_bar: str
            The name of the kernel function for Nadaraya-Watson conditional CDF 
            estimator. (Default: "gaussian".)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. (Default: h=None, b=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used
            with additional scaling factors C_h and C_b, respectively.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome function. 
            (Default: deriv_ord=1. Then, it estimates the partial derivative of the 
            conditional mean outcome function with respect to the treatment variable.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
        parallel: boolean
            The indicator of whether the function should be parallel executed by
            multi-processing. (Default: parallel=False.)
            
        processes: int
            The number of processes for parallel execution. (Default: processes=20.)
    
    Return
    ----------
        theta_C: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
    '''
    
    if t_eval is None: 
        t_eval = X[:,0].copy()
        
    n = X.shape[0]  ## Number of data points
    d = 1
    if h_bar is None:
        # Apply the Silverman's rule of thumb bandwidth in Chen et al.(2016).
        h_bar = (4/(d+2))**(1/(d+4))*(n**(-1/(d+4)))*np.std(X[:,0])
    if print_bw:
        print("The current bandwidth for the conditional CDF estimator is "+ str(h_bar) + ".\n")
    
    kernT_bar, sigmaK_sq, K_sq = KernelRetrieval(kernT_bar)
        
    weight_mat = kernT_bar((t_eval - X[:,0].reshape(-1,1)) / h_bar)
    weight_mat = weight_mat / np.sum(weight_mat, axis=0)
    weight_mat[np.isnan(weight_mat)] = 0
    if parallel:
        with Pool(processes=processes) as pool:
            part_fun = partial(LocalPolyReg_Fs, Y=Y, X=X, degree=degree, deriv_ord=deriv_ord, 
                               h=h, b=b, C_h=C_h, C_b=C_b, print_bw=print_bw, kernT=kernT, kernS=kernS)
            beta_mat = pool.map(part_fun, [np.concatenate([t_eval[i]*np.ones((n,1)), X[:,1:]], axis=1) for i in range(t_eval.shape[0])])
            beta_mat = np.concatenate(beta_mat, axis=0).reshape(t_eval.shape[0], n).T
    else:
        beta_mat = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            X_mat = np.concatenate([t_eval[i]*np.ones((n,1)), X[:,1:]], axis=1)
            beta_mat[:,i] = LocalPolyReg(Y, X, x_eval=X_mat, degree=degree, deriv_ord=deriv_ord, 
                                         h=h, b=b, C_h=C_h, C_b=C_b, print_bw=print_bw, 
                                         kernT=kernT, kernS=kernS)
    
    theta_C = np.sum(weight_mat * beta_mat, axis=0)
    return theta_C


def DerivEffectBoot(Y, X, t_eval=None, boot_num=500, alpha=0.95, h_bar=None, 
                    kernT_bar="gaussian", h=None, b=None, C_h=7, C_b=3, 
                    print_bw=True, degree=2, deriv_ord=1, 
                    kernT="epanechnikov", kernS="epanechnikov", 
                    parallel=False, processes=20):
    '''
    Conduct inference on the derivative of the dose-response curve via Nadaraya-Watson 
    conditional CDF estimator and nonparametric bootstrap.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points. (Default: t_eval=None. 
            Then, t_eval=X[:,0], which consists of the observed treatment variables.)
            
        boot_num: int
            The number of bootstrapping times. (Default: bootstrap_num=500.) 
            
        alpha: float
            The confidence level of both the uniform confidence band and pointwise
            confidence interval. (Default: alpha=0.95.)
            
        h_bar: float
            The bandwidth parameters for the Nadaraya-Watson conditional CDF estimator. 
            (Default: h_bar=None. Then, the Silverman's rule of thumb is applied. 
            See Chen et al.(2016) for details.)
            
        kernT_bar: str
            The name of the kernel function for Nadaraya-Watson conditional CDF 
            estimator. (Default: "gaussian".)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. (Default: h=None, b=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used
            with additional scaling factors C_h and C_b, respectively.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome function. 
            (Default: deriv_ord=1. Then, it estimates the partial derivative of the 
            conditional mean outcome function with respect to the treatment variable.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
        parallel: boolean
            The indicator of whether the function should be parallel executed by
            multi-processing. (Default: parallel=False.)
            
        processes: int
            The number of processes for parallel execution. (Default: processes=20.)
    
    Return
    ----------
        theta_C: (m,)-array
            The estimated derivative of the dose-response curve evaluated at 
            points "t_eval".
            
        theta_C_boot: (m,)-array
            The estimated derivatives of the dose-response curve on bootstrap samples 
            evaluated at points "t_eval".
            
        theta_alpha: float
            The width of the uniform confidence band.
            
        theta_alpha_var: (m,)-array
            The widths of the pointwise confidence bands at evaluation points "t_eval".
    '''
    
    if t_eval is None: 
        t_eval = X[:,0].copy()
        
    n = X.shape[0]  ## Number of data points
    
    theta_est = DerivEffect(Y, X, t_eval=t_eval, h_bar=h_bar, kernT_bar=kernT_bar, 
                            h=h, b=b, C_h=C_h, C_b=C_b, print_bw=print_bw, 
                            degree=degree, deriv_ord=deriv_ord, 
                            kernT=kernT, kernS=kernS, parallel=parallel, processes=processes)
    
    theta_C_boot = np.zeros((boot_num, t_eval.shape[0]))
    b = 0
    while b < boot_num:
        ind = np.random.choice(n, size=n, replace=True)
        X_boot = X[ind,:]
        Y_boot = Y[ind]
        theta_C_boot[b,:] = DerivEffect(Y_boot, X_boot, t_eval=t_eval, h_bar=h_bar, 
                                   kernT_bar=kernT_bar, h=h, b=b, C_h=C_h, C_b=C_b, 
                                   print_bw=print_bw, degree=degree, deriv_ord=deriv_ord, 
                                   kernT=kernT, kernS=kernS, parallel=parallel, 
                                   processes=processes)
        if np.sum(np.isnan(theta_C_boot[b,:])) == 0:
            b += 1
    
    # Compute the alpha% uniform confidence bands
    theta_boot_sup = np.max(np.abs(theta_C_boot - theta_est), axis=1)
    theta_alpha = np.quantile(theta_boot_sup, alpha)

    # Compute the alpha% pointwise confidence intervals
    theta_boot_abs = np.abs(theta_C_boot - theta_est)
    theta_alpha_var = np.quantile(theta_boot_abs, alpha, axis=0)
    
    return theta_est, theta_C_boot, theta_alpha, theta_alpha_var


def IntegEst(Y, X, t_eval=None, h_bar=None, kernT_bar="gaussian", h=None, b=None, 
             C_h=7, C_b=3, print_bw=True, degree=2, deriv_ord=1, kernT="epanechnikov", 
             kernS="epanechnikov", parallel=False, processes=20):
    '''
    Estimating the dose-response curve via our integral estimator with linear 
    interpolation approximation.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points. (Default: t_eval=None. 
            Then, t_eval=X[:,0].)
            
        h_bar: float
            The bandwidth parameters for the Nadaraya-Watson conditional CDF estimator. 
            (Default: h_bar=None. Then, the Silverman's rule of thumb is applied. 
            See Chen et al.(2016) for details.)
            
        kernT_bar: str
            The name of the kernel function for Nadaraya-Watson conditional CDF estimator.
            (Default: "gaussian".)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. (Default: h=None, b=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used
            with additional scaling factors C_h and C_b, respectively.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome function. 
            (Default: deriv_ord=1. Then, it estimates the partial derivative of the 
            conditional mean outcome function with respect to the treatment variable.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
        
        parallel: boolean
            The indicator of whether the function should be parallel executed by
            multi-processing. (Default: parallel=False.)
            
        processes: int
            The number of processes for parallel execution. (Default: processes=20.)
            
    Return
    ----------
        m_est: (m,)-array
            The estimated dose-response curve evaluated at points "t_eval".
    '''
    
    if t_eval is None: 
        t_eval = X[:,0].copy()
    
    T_sort = np.sort(X[:,0])
    n = X.shape[0]  ## Number of data points
    
    # Compute theta_C at the order statistics of T
    theta_est = DerivEffect(Y, X, t_eval=T_sort, h_bar=h_bar, kernT_bar=kernT_bar, 
                            h=h, b=b, C_h=C_h, C_b=C_b, print_bw=print_bw, 
                            degree=degree, deriv_ord=deriv_ord, kernT=kernT, 
                            kernS=kernS, parallel=parallel, processes=processes)
    T_delta = T_sort[1:] - T_sort[:(n-1)]
    
    int_mat_up = np.ones((n,)) * (T_delta*(np.arange(1, n)*theta_est[:(n-1)])).reshape(-1,1)
    int_mat_up = int_mat_up * (np.arange(n-1).reshape(-1,1) < np.arange(n))
    
    int_mat_down = np.ones((n,)) * (T_delta*((n-np.arange(1, n))*theta_est[1:])).reshape(-1,1)
    int_mat_down = int_mat_down * (np.arange(n-1).reshape(-1,1) >= np.arange(n))
    m_samp = np.mean(Y) + np.sum(int_mat_up - int_mat_down, axis=0)/n
    
    m_est = np.interp(t_eval, T_sort, m_samp)
            
    return m_est


def IntegEstBoot(Y, X, t_eval=None, boot_num=500, alpha=0.95, h_bar=None, 
                 kernT_bar="gaussian", h=None, b=None, C_h=7, C_b=3, print_bw=True, 
                 degree=2, deriv_ord=1, kernT="epanechnikov", kernS="epanechnikov", 
                 parallel=False, processes=20):
    '''
    Conduct inference on the dose-response curve via our integral estimator and
    nonparametric bootstrap.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points. (Default: t_eval=None. 
            Then, t_eval=X[:,0].)
            
        boot_num: int
            The number of bootstrapping times. (Default: bootstrap_num=500.) 
        
        alpha: float
            The confidence level of both the uniform confidence band and pointwise
            confidence interval. (Default: alpha=0.95.)
            
        h_bar: float
            The bandwidth parameters for the Nadaraya-Watson conditional CDF estimator. 
            (Default: h_bar=None. Then, the Silverman's rule of thumb is applied. 
            See Chen et al.(2016) for details.)
            
        kernT_bar: str
            The name of the kernel function for Nadaraya-Watson conditional CDF estimator.
            (Default: "gaussian".)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. (Default: h=None, b=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used
            with additional scaling factors C_h and C_b, respectively.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative the conditional mean outcome function. 
            (Default: deriv_ord=1. Then, it estimates the partial derivative of the 
            conditional mean outcome function with respect to the treatment variable.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
        
        parallel: boolean
            The indicator of whether the function should be parallel executed by
            multi-processing. (Default: parallel=False.)
            
        processes: int
            The number of processes for parallel execution. (Default: processes=20.)
            
    Return
    ----------
        m_est: (m,)-array
            The estimated dose-response curve evaluated at points "t_eval".
            
        m_est_boot: (boot_num, m)-array
            The estimated dose-response curves (or their derivatives) on the bootstrap 
            samples evaluated at points "t_eval".
            
        m_alpha: float
            The width of the uniform confidence band.
            
        m_alpha_var: (m,)-array
            The widths of the pointwise confidence bands at evaluation points "t_eval".
    '''
    
    if t_eval is None: 
        t_eval = X[:,0].copy()
        
    m_est = IntegEst(Y, X, t_eval=t_eval, h_bar=h_bar, kernT_bar=kernT_bar, 
                     h=h, b=b, C_h=C_h, C_b=C_b, print_bw=print_bw, degree=degree, 
                     deriv_ord=deriv_ord, kernT=kernT, kernS=kernS, parallel=parallel, 
                     processes=processes)
    
    n = X.shape[0]  ## Number of data points
    m_est_boot = np.zeros((boot_num, t_eval.shape[0]))
    b = 0
    while b < boot_num:
        ind = np.random.choice(n, size=n, replace=True)
        X_boot = X[ind,:]
        Y_boot = Y[ind]
        m_est_boot[b,:] = IntegEst(Y_boot, X_boot, t_eval=t_eval, h_bar=h_bar, 
                                   kernT_bar=kernT_bar, h=h, b=b, C_h=C_h, C_b=C_b, 
                                   print_bw=print_bw, degree=degree, deriv_ord=deriv_ord, 
                                   kernT=kernT, kernS=kernS, parallel=parallel, 
                                   processes=processes)
        if np.sum(np.isnan(m_est_boot[b,:])) == 0:
            b += 1
            
    # Compute the alpha% uniform confidence bands
    m_boot_sup = np.max(np.abs(m_est_boot - m_est), axis=1)
    m_alpha = np.quantile(m_boot_sup, alpha)

    # Compute the alpha% pointwise confidence intervals
    m_boot_abs = np.abs(m_est_boot - m_est)
    m_alpha_var = np.quantile(m_boot_abs, alpha, axis=0)
    return m_est, m_est_boot, m_alpha, m_alpha_var


def LocalPolyReg1D(Y, X, h=None, x_eval=None, degree=2, deriv_ord=0, kernel="epanechnikov"):
    '''
    Local polynomial regression in one dimension.
    
    Parameters
    ----------
        Y: (m,)-array
            The y coordinates of m data points.
        
        X: (m,)-array
            The x coordinates of m data points.
            
        h: float
            The bandwidth parameter. (Default: h=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used.)
            
        x_eval: (k,)-array
            Vector of evaluation points. (Default: x_eval=None. Then, x_eval=X.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of derivatives of the regression function that are estimated. 
            (Default: deriv_ord=0. Then, it is the usual local polynomial regression.)
            
    Return
    ----------
        Y_est: (m,)-array
            The estimated function or its derivatives by local polynomial regression
            evaluated at points "x_eval".
    '''
    if x_eval is None: 
        x_eval = X.copy()
        
    n = X.shape[0]  ## Number of data points
        
    kernel, sigmaK_sq, K_sq = KernelRetrieval(kernel)
    if h is None:
        # Apply the rule-of-thumb bandwidth selector in Eq.(A1) of Yang and Tschernig (1999)
        p_coeff = np.polyfit(X, Y, 4)
        sec_deriv = 12*p_coeff[0]*X + 6*p_coeff[1]*X + 2*X
        C_fun = np.mean(sec_deriv**2)
        lhs = np.concatenate([np.ones((n,1)), X.reshape(-1,1)], axis=1)
        rcond = np.finfo(lhs.dtype).eps * max(*lhs.shape)
        beta = np.linalg.lstsq(lhs, Y, rcond=rcond)[0]
        # Compute the integrated residual sum of squares
        resid = np.sum((Y - np.dot(lhs, beta))**2) * (np.max(X) - np.min(X))/ (n-5)
        sigmaK_sq = sigmaK_sq**2
        # ROT
        h = ((K_sq*resid)/(4*n*sigmaK_sq*C_fun))**(1/5) * 7
        # h = ((K_sq*resid)/(4*n*sigmaK_sq*C_fun))**(1/5) * (n**(d/(5*(d+5)))) * 5
    print("The current bandwidth for the local polynomial regression is "+ str(h) + ".\n")
    
    Y_est = np.zeros_like(x_eval)
    for i in range(x_eval.shape[0]):
        weights = kernel((X-x_eval[i])/h)
        # Filter out the data points with zero weights to speed up regressions with kernels of local support.
        inds = np.where(np.abs(weights)>1e-26)[0]
        X_dat = np.zeros((n, degree+1))
        for p in range(degree + 1):
            X_dat[:,p] = (X - x_eval[i])**p
        S = np.sqrt(weights)[inds]
        lhs = np.dot(np.diag(S), X_dat[inds,:])
        rhs = S*Y[inds]
        rcond = np.finfo(lhs.dtype).eps * max(*lhs.shape)
        beta = np.linalg.lstsq(lhs, rhs, rcond=rcond)[0]
        Y_est[i] = np.math.factorial(deriv_ord)*beta[deriv_ord]
    return Y_est



def RegAdjust(Y, X, t_eval=None, h=None, b=None, C_h=7, C_b=3, print_bw=True, 
              degree=2, deriv_ord=0, kernT="epanechnikov", kernS="epanechnikov", 
              parallel=False, processes=20):
    '''
    Estimating the dose-response curve via simple integral estimator with linear 
    interpolation approximation.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcomes of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points. (Default: t_eval=None. 
            Then, t_eval=X[:,0].)
            
        h,b: float
            The bandwidth parameters for the treatment/exposure variable and 
            confounding variables. (Default: h=None, b=None. Then, the rule-of-thumb 
            bandwidth selector in Eq.(A1) of Yang and Tschernig (1999) is used
            with additional scaling factors C_h and C_b, respectively.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
            
        degree: int
            Degree of local polynomials. (Default: degree=2.)
            
        deriv_ord: int
            The order of the estimated derivative of the conditional mean outcome 
            function. (Default: deriv_ord=0. Then, it estimates the conditional 
            mean outcome function itself.)
            
        kernT, kernS: str
            The names of kernel functions for the treatment/exposure variable 
            and confounding variables. (Default: "epanechnikov".)
            
        parallel: boolean
            The indicator of whether the function should be parallel executed by
            multi-processing. (Default: parallel=False.)
            
        processes: int
            The number of processes for parallel execution. (Default: processes=20.)
            
    Return
    ----------
        m_est: (m,)-array
            The estimated dose-response curve (or its derivative) evaluated 
            at points "t_eval".
    '''
    
    if t_eval is None: 
        t_eval = X[:,0].copy()
    
    n = X.shape[0]  ## Number of data points
    
    if parallel:
        with Pool(processes=processes) as pool:
            part_fun = partial(LocalPolyReg_Fs, Y=Y, X=X, degree=degree, 
                               deriv_ord=deriv_ord, h=h, b=b, C_h=C_h, C_b=C_b, 
                               print_bw=print_bw, kernT=kernT, kernS=kernS)
            beta_mat = pool.map(part_fun, [np.concatenate([t_eval[i]*np.ones((n,1)), X[:,1:]], axis=1) for i in range(t_eval.shape[0])])
            beta_mat = np.concatenate(beta_mat, axis=0).reshape(t_eval.shape[0], n).T
    else:
        beta_mat = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            X_mat = np.concatenate([t_eval[i]*np.ones((n,1)), X[:,1:]], axis=1)
            beta_mat[:,i] = LocalPolyReg(Y, X, x_eval=X_mat, degree=degree, 
                                         deriv_ord=deriv_ord, h=h, b=b, C_h=C_h, 
                                         C_b=C_b, print_bw=print_bw, kernT=kernT, kernS=kernS)
    m_est = np.mean(beta_mat, axis=0)
    
    return m_est




# -*- coding: utf-8 -*-

# Author: Yikun Zhang
# Last Editing: Dec 22, 2024

# Description: This script contains the implementations of the regression 
# adjustment (RA), inverse probability weighting (IPW), and doubly robust (DR) 
# estimators of the dose-response curve under the positivity condition.

import numpy as np
from .rbf import KernelRetrieval
from .utils import CondDenEst, CondDenEstKDE
from sklearn.model_selection import KFold

#=======================================================================================#

def RegAdjustDR(Y, X, t_eval, mu, L=1, multi_boot=False, B=1000):
    '''
    Estimating the dose-response curve through the regression adjustment
    (or G-computation) form.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        mu: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        multi_boot: boolean
            An indicator of whether the multiplier bootstrap will be run. 
            (Default: multi_boot=False.)
            
        B: int
            The number of bootstrapping times. (Default: B=1000.)
            
    Return
    ----------
        m_est: (m,)-array
            The estimated dose-response curve evaluated at points "t_eval".
            
        mu_boot: (B,m)-array
            The estimated dose-response curves on bootstrapping data evaluated
            at points "t_eval". (Only return this quantity when "multi_boot=True".)
    '''
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the regression model on the entire data
        mu_hat = mu.fit(X, Y)
        mu_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the fitted regression model
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            mu_est[:,i] = mu_hat.predict(X_eval)
    else:
        # Conduct L-fold cross-fittings: fit the regression model on the training fold data
        # and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        mu_est = np.zeros((n, t_eval.shape[0]))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            
            mu_hat = mu.fit(X_tr, Y_tr)
            for i in range(t_eval.shape[0]):
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                mu_est[te_ind,i] = mu_hat.predict(X_eval_te)
    
    if multi_boot:
        mu_boot = np.zeros((B, t_eval.shape[0]))
        for b in range(B):
            Z = np.random.randn(n, t_eval.shape[0]) + 1
            mu_boot[b,:] = np.mean(Z * mu_est, axis=0)
        m_est = np.mean(mu_est, axis=0)
        return m_est, mu_boot
    else:
        m_est = np.mean(mu_est, axis=0)
        return m_est


def IPWDR(Y, X, t_eval, condTS_type, condTS_mod, L, h, kern="epanechnikov", 
          tau=0.001, b=None, self_norm=True):
    '''
    Estimating the dose-response curve through the inverse probability weighting
    (IPW) form.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        condTS_type: str
            Specifying the model type for estimating the conditional density of
            the treatment variable T given the covariate vector S.
            
        condTS_mod: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The regression model for estimating the conditional density of T given S.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            (Default: L=1.)
            
        h: float
            The bandwidth parameter.
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        tau: float
            The threshold value that lower bounds the estimated conditional density
            values. (Default: tau=0.001.)
            
        b: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
    Return
    ----------
        m_est: (m,)-array
            The estimated dose-response curve evaluated at points "t_eval".
            
        cond_est_full: (n,)-array
            The estimated conditional density function of T given S evaluated at
            the n observed data points.
    '''
    
    kern_type = kern
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern_type)
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the conditional density model on the entire data
        if condTS_type == 'true':
            condTS_est = condTS_mod
        elif condTS_type == 'kde':
            condTS_est = CondDenEstKDE(X[:,0], X[:,1:], reg_mod=condTS_mod, y_eval=X[:,0], 
                                       x_eval=X[:,1:], kern=kern_type, b=b)
        else:
            condTS_est = CondDenEst(X[:,0], X[:,1:], reg_mod=condTS_mod, y_eval=X[:,0], 
                                    x_eval=X[:,1:], kern='gaussian', b=b)
        condTS_est[condTS_est < tau] = tau
        
        m_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for i in range(t_eval.shape[0]):
            # Self-normalizing weights
            norm_w[i] = np.sum(kern((t_eval[i] - X[:,0])/h) / condTS_est) / h
            m_hat[:,i] = kern((t_eval[i] - X[:,0])/h) * Y / (h * condTS_est)

        if self_norm:
            # Self-normalized IPW estimator
            m_hat = m_hat / norm_w
            m_est = np.sum(m_hat, axis=0, where=~np.isnan(m_hat))
        else:
            m_est = np.mean(m_hat, axis=0, where=~np.isnan(m_hat))
        cond_est_full = condTS_est.copy()
    else:
        # Conduct L-fold cross-fittings: fit the conditional density model on the training fold 
        # data and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        m_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        cond_est_full = np.zeros((n,))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            X_te = X[te_ind,:]
            Y_te = Y[te_ind]
            
            if condTS_type == 'true':
                condTS_est = condTS_mod[te_ind]
            elif condTS_type == 'kde':
                condTS_est = CondDenEstKDE(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                           y_eval=X_te[:,0], x_eval=X_te[:,1:], kern=kern_type, b=b)
            else:
                condTS_est = CondDenEst(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                        y_eval=X_te[:,0], x_eval=X_te[:,1:], kern='gaussian', b=b)
            condTS_est[condTS_est < tau] = tau
            cond_est_full[te_ind] = condTS_est
            for i in range(t_eval.shape[0]):
                # Self-normalizing weights
                w = np.sum(kern((t_eval[i] - X[te_ind,0])/h) / condTS_est) / h
                norm_w[i] = norm_w[i] + w
                m_hat[te_ind,i] = kern((t_eval[i] - X[te_ind,0])/h) * Y_te / (h * condTS_est)

        if self_norm:
            norm_w[norm_w == 0] = 1
            m_est = np.sum(m_hat, axis=0, where=~np.isnan(m_hat)) / norm_w
        else:
            m_est = np.mean(m_hat, axis=0, where=~np.isnan(m_hat))
    return m_est, cond_est_full



def DRDR(Y, X, t_eval, mu, condTS_type, condTS_mod, L, h, kern="epanechnikov", 
         tau=0.001, b=None, self_norm=True):
    '''
    Estimating the dose-response curve through the doubly robust (DR) form.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are the confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points.
            
        mu: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        condTS_type: str
            Specifying the model type for estimating the conditional density of
            the treatment variable T given the covariate vector S.
            
        condTS_mod: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The regression model for estimating the conditional density of T given S.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            
        h: float
            The bandwidth parameter.
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        tau: float
            The threshold value that lower bounds the estimated conditional density
            values. (Default: tau=0.001.)
            
        b: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
    Return
    ----------
        m_est: (m,)-array
            The estimated dose-response curve evaluated at points "t_eval".
            
        sd_est: (m,)-array
            The estimated asymptotic stdndard deviation of the DR estimator 
            evaluated at points "t_eval".
    '''
    
    kern_type = kern
    kern, sigmaK_sq, K_sq = KernelRetrieval(kern)
    n = X.shape[0]  ## Number of data points
    if L <= 1:
        # No cross-fittings: fit the conditional density model and the regression model on the entire data
        if condTS_type == 'true':
            condTS_est = condTS_mod
        elif condTS_type == 'kde':
            condTS_est = CondDenEstKDE(X[:,0], X[:,1:], reg_mod=condTS_mod, 
                                       y_eval=X[:,0], x_eval=X[:,1:], kern=kern_type, b=b)
        else:
            condTS_est = CondDenEst(X[:,0], X[:,1:], reg_mod=condTS_mod, y_eval=X[:,0], 
                                    x_eval=X[:,1:], kern='gaussian', b=b)
        condTS_est[condTS_est < tau] = tau
        mu_fit = mu.fit(X, Y)
        mu_hat = np.zeros((n, t_eval.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        for i in range(t_eval.shape[0]):
            # Define the data matrix for evaluating the fitted regression model
            X_eval = np.column_stack([t_eval[i]*np.ones(n), X[:,1:]])
            mu_hat[:,i] = mu_fit.predict(X_eval)
            IPW_hat[:,i] = kern((t_eval[i] - X[:,0])/h) * (Y - mu_hat[:,i]) / (h * condTS_est)
            # Self-normalizing weights
            norm_w[i] = np.sum(kern((t_eval[i] - X[:,0])/h) / condTS_est) / (n * h)
            
        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        m_hat = IPW_hat + mu_hat
        m_est = np.mean(m_hat, axis=0, where=~np.isnan(m_hat))
        
        # Estimate the variance of m(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (mu_hat[:,i] - m_est[i]))**2 * h
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*h))
    else:
        # Conduct L-fold cross-fittings: fit the reciprocal of the conditional model 
        # and the regression model on the training fold data and evaluate it on the test fold data
        kf = KFold(n_splits=L, shuffle=True, random_state=0)
        mu_hat = np.zeros((n, t_eval.shape[0]))
        IPW_hat = np.zeros((n, t_eval.shape[0]))
        norm_w = np.zeros((t_eval.shape[0],))
        cond_est_full = np.zeros((n,))
        for tr_ind, te_ind in kf.split(X):
            X_tr = X[tr_ind,:]
            Y_tr = Y[tr_ind]
            X_te = X[te_ind,:]
            Y_te = Y[te_ind]
            
            if condTS_type == 'true':
                condTS_est = condTS_mod[te_ind]
            elif condTS_type == 'kde':
                condTS_est = CondDenEstKDE(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                           y_eval=X_te[:,0], x_eval=X_te[:,1:], kern=kern_type, b=b)
            else:
                condTS_est = CondDenEst(X_tr[:,0], X_tr[:,1:], reg_mod=condTS_mod, 
                                        y_eval=X_te[:,0], x_eval=X_te[:,1:], kern='gaussian', b=b)
            condTS_est[condTS_est < tau] = tau
            cond_est_full[te_ind] = condTS_est
            mu_fit = mu.fit(X_tr, Y_tr)
            for i in range(t_eval.shape[0]):
                X_eval_te = np.column_stack([t_eval[i]*np.ones(X_te.shape[0]), X_te[:,1:]])
                mu_hat[te_ind,i] = mu_fit.predict(X_eval_te)
                IPW_hat[te_ind,i] = kern((t_eval[i] - X[te_ind,0])/h) * (Y_te - mu_hat[te_ind,i]) / (h * condTS_est)
                
                # Self-normalizing weights
                w = np.sum(kern((t_eval[i] - X[te_ind,0])/h) / condTS_est) / (n * h)
                norm_w[i] = norm_w[i] + w
        
        if self_norm:
            IPW_hat = IPW_hat / norm_w
        # Add up the IPW and RA components
        m_hat = IPW_hat + mu_hat
        m_est = np.mean(m_hat, axis=0, where=~np.isnan(m_hat))
        
        # Estimate the variance of m(t) using the square of the influence function
        var_est = np.zeros((n, t_eval.shape[0]))
        for i in range(t_eval.shape[0]):
            var_est[:,i] = (IPW_hat[:,i] + (mu_hat[:,i] - m_est[i]))**2 * h
        sd_est = np.sqrt(np.mean(var_est, axis=0)/(n*h))
    return m_est, sd_est


def DRCurve(Y, X, t_eval=None, est="RA", mu=None, condTS_type=None, condTS_mod=None, 
            L=1, h=None, kern="epanechnikov", tau=0.001, h_cond=None, self_norm=True, 
            print_bw=True):
    '''
    Dose-response curve estimation under the positivity condition.
    
    Parameters
    ----------
        Y: (n,)-array
            The outcome variables of n observations.
            
        X: (n,d+1)-array
            The first column of X is the treatment/exposure variable, while 
            the other d columns are confounding variables of n observations.
            
        t_eval: (m,)-array
            The coordinates of the m evaluation points. (Default: t_eval=None. 
            Then, t_eval=X[:,0], which consists of the observed treatment variables.)
            
        est: str
            The type of the dose-response curve estimator. (Default: est="RA". 
            Other choices include "IPW" and "DR".)
            
        mu: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The conditional mean outcome (or regression) model of Y given X.
            
        condTS_type: str
            Specifying the model type for estimating the conditional density of
            the treatment variable T given the covariate vector S.
            
        condTS_mod: scikit-learn model or any python model that can use ".fit()" and ".predict()"
            The regression model for estimating the conditional density of T given S.
            
        L: int
            The number of data folds for cross-fitting. When L<= 1, no cross-fittings 
            are applied and the regression model is fitted on the entire dataset.
            
        h: float
            The bandwidth parameter for the IPW/DR estimator. (Default: h=None. 
            Then the Silverman's rule of thumb is applied; see Chen et al.(2016) 
            for details.)
            
        kern: str
            The name of the kernel function. (Default: kern="epanechnikov".)
            
        tau: float
            The threshold value that lower bounds the estimated conditional density
            values. (Default: tau=0.001.)
            
        h_cond: float
            The bandwidth parameter for the kernel-smoothed conditional density
            estimation methods. (Default: b=None.)
            
        self_norm: boolean
            An indicator of whether the self-normalized version is implemented.
            (Default: self_norm=True.)
            
        print_bw: boolean
            The indicator of whether the current bandwidth parameters should be
            printed to the console. (Default: print_bw=True.)
    
    Return
    ----------
        m_est: (m,)-array
            The estimated dose-response curve evaluated at points "t_eval".
            
        sd_est: (m,)-array (if est="DR")
            The estimated asymptotic standard deviation of the DR estimator 
            evaluated at points "t_eval".
    '''
    
    if t_eval is None: 
        t_eval = X[:,0].copy()
    
    n = X.shape[0]  ## Number of data points
    
    if h is None:
        # Apply Silverman's rule of thumb to select the bandwidth parameter 
        h = (4/3)**(1/5)*(n**(-1/5))*np.std(X[:,0])
    
    if print_bw:
        print("The current bandwidth for the "+str(est)+" estimator is "+ str(h) + ".\n")
    
    if est == "RA":
        m_est = RegAdjustDR(Y, X, t_eval, mu, L)
    elif est == "IPW":
        m_est, cond_est = IPWDR(Y, X, t_eval, condTS_type, condTS_mod, L, h, kern, tau, h_cond, 
                                self_norm)
    else:
        m_est = DRDR(Y, X, t_eval, mu, condTS_type, condTS_mod, L, h, kern, tau, h_cond, self_norm)
    
    return m_est

